extern crate object;

use std::fs::File;
use std::io::prelude::*;
use std::path::Path;
use std::path::PathBuf;

use clap::ValueEnum;
use log::debug;
use log::warn;
use object::elf;
use object::Endianness;

use ihex;
use object::write::elf::{FileHeader, SectionHeader};

// Perhaps should refactor so we don't have UI ValueEnum here?
#[derive(Debug, Clone, ValueEnum)]
pub enum MemoryFileType {
    Elf,
    Ihex,
    Bin,
}

const M33_FLAGS: u32 = elf::EF_ARM_EABI_VER5 | elf::EF_ARM_ABI_FLOAT_HARD; // processor has hard float, even if code may not use it

struct ElfWriter;
struct IhexWriter;
struct BinWriter;

struct ElfReader;
struct IhexReader;
struct BinReader {
    filesize: usize,
    file: File,
}

/// Build a new object to manage writing type T
impl MemoryFileType {
    fn type_from_filename(filename: &str) -> Option<MemoryFileType> {
        let filename = filename.to_lowercase();
        let (_, extension) = filename.rsplit_once('.').unwrap_or(("", ""));
        match extension {
            "elf" => Some(MemoryFileType::Elf),
            "hex" | "ihex" => Some(MemoryFileType::Ihex),
            "bin" | "raw" => Some(MemoryFileType::Bin),
            _ => None,
        }
    }
    pub fn read_object_file(
        filename: &PathBuf,
    ) -> Result<Box<dyn ObjectFileReader>, std::io::Error> {
        match MemoryFileType::type_from_filename(filename.to_str().unwrap_or("")) {
            Some(MemoryFileType::Elf) => unimplemented!(),
            Some(MemoryFileType::Ihex) => unimplemented!(),
            Some(MemoryFileType::Bin) => BinReader::new(filename),
            None => {
                warn!("Can't determine an appropriate output type based on filename {}, reading as raw binary", filename.to_str().unwrap_or("<invalid utf-8>"));
                unimplemented!()
            }
        }
    }
    pub fn mem_writer(&self) -> Box<dyn MemoryWriter> {
        match *self {
            MemoryFileType::Elf => Box::new(ElfWriter),
            MemoryFileType::Ihex => Box::new(IhexWriter),
            MemoryFileType::Bin => Box::new(BinWriter),
        }
    }
    pub fn mem_writer_for_file(filename: &PathBuf) -> Box<dyn MemoryWriter> {
        match MemoryFileType::type_from_filename(filename.to_str().unwrap_or("")) {
            Some(MemoryFileType::Elf) => Box::new(ElfWriter),
            Some(MemoryFileType::Ihex) => Box::new(IhexWriter),
            Some(MemoryFileType::Bin) => Box::new(BinWriter),
            None => {
                warn!("Can't determine an appropriate output type based on filename {}, writing as raw binary", filename.to_str().unwrap_or("<invalid utf-8>"));
                Box::new(BinWriter)
            }
        }
    }
}

/// An object that can read a block of memory from an object file
pub trait ObjectFileReader {
    fn new(file: &PathBuf) -> Result<Box<dyn ObjectFileReader>, std::io::Error>
    where
        Self: Sized;
    fn len(&self) -> usize;
    /// Set buf's contents to the object file's first `text` section, up to `size` bytes
    fn read_first(
        &mut self,
        buf: &mut Vec<u8>,
        size: Option<u32>,
    ) -> Result<(), Box<dyn std::error::Error>>;
    /// Set buf's contents to the object file's contents, starting from LMA `addr`` to a maximum of `size` bytes,
    /// or to the end of the section containing `addr`
    fn read_at(
        &mut self,
        buf: &mut Vec<u8>,
        addr: u32,
        size: Option<u32>,
    ) -> Result<(), Box<dyn std::error::Error>>;
}

/// An object that can write a block of memory to a file
pub trait MemoryWriter {
    fn write_mem(
        &mut self,
        output: &mut dyn std::io::Write,
        addr: u32,
        data: &[u8],
    ) -> Result<(), Box<dyn std::error::Error>>;
}

impl MemoryWriter for ElfWriter {
    fn write_mem(
        &mut self,
        output: &mut dyn std::io::Write,
        addr: u32,
        data: &[u8],
    ) -> Result<(), Box<dyn std::error::Error>> {
        debug!("Writing as ELF with base address 0x{:x}", addr);
        let mut buf = Vec::new();

        // object is not well documented, see how this API must be used here:
        // https://github.com/gimli-rs/object/blob/master/src/write/elf/object.rs#L217

        let mut writer = object::write::elf::Writer::new(Endianness::Little, false, &mut buf);

        // Calculating offsets

        writer.reserve_file_header();

        let _text_id = writer.reserve_section_index();
        let data_ofs = writer.reserve(data.len(), 4);
        let text_name = writer.add_section_name(".text".as_bytes());

        writer.reserve_null_symbol_index();
        let symtab_num_local = writer.symbol_count();
        writer.reserve_symtab_section_index();
        writer.reserve_symtab();
        writer.reserve_symtab_shndx();
        writer.reserve_strtab_section_index();
        writer.reserve_strtab();
        writer.reserve_shstrtab_section_index();
        writer.reserve_shstrtab();
        writer.reserve_section_headers();

        // Writing

        writer.write_file_header(&FileHeader {
            abi_version: elf::EV_CURRENT,
            os_abi: elf::ELFOSABI_NONE,
            e_type: elf::ET_EXEC,
            e_machine: elf::EM_ARM,
            e_entry: addr.into(),
            e_flags: M33_FLAGS,
        })?;
        writer.write_align(4);
        writer.write(&data);
        writer.write_null_symbol();
        writer.write_symtab_shndx();
        writer.write_strtab();

        writer.write_shstrtab();
        writer.write_null_section_header();

        writer.write_section_header(&SectionHeader {
            name: Some(text_name),
            sh_type: elf::SHT_PROGBITS,
            sh_flags: (elf::SHF_ALLOC | elf::SHF_EXECINSTR).into(),
            sh_addr: addr.into(),
            sh_offset: data_ofs as u64,
            sh_size: data.len() as u64,
            sh_link: elf::SHN_UNDEF.into(),
            sh_info: 0,
            sh_addralign: 4,
            sh_entsize: 0,
        });

        writer.write_symtab_section_header(symtab_num_local);
        writer.write_symtab_shndx_section_header();
        writer.write_strtab_section_header();
        writer.write_shstrtab_section_header();

        output.write_all(&buf)?;

        Ok(())
        // let triple = target_lexicon::Triple {
        //     architecture: Architecture::Arm(target_lexicon::ArmArchitecture::Armv8mBase),
        //     vendor: Vendor::Unknown,
        //     operating_system: OperatingSystem::None_,
        //     environment: Environment::Eabihf,
        //     binary_format: BinaryFormat::Elf,
        // };
    }
}

impl MemoryWriter for IhexWriter {
    fn write_mem(
        &mut self,
        output: &mut dyn std::io::Write,
        addr: u32,
        data: &[u8],
    ) -> Result<(), Box<dyn std::error::Error>> {
        let mut records = Vec::new();

        const BLOCK_SIZE: u32 = u16::MAX as u32 + 1;
        const CHUNK_SIZE: u32 = 16;

        for (block, data) in data.chunks(BLOCK_SIZE as usize).enumerate() {
            records.push(ihex::Record::ExtendedLinearAddress(
                (((addr + block as u32 * BLOCK_SIZE) & 0xffff0000) >> 16)
                    .try_into()
                    .unwrap(),
            ));
            records.extend(
                data.chunks(CHUNK_SIZE as usize)
                    .enumerate()
                    .map(|(offset, chunk)| ihex::Record::Data {
                        offset: ((offset as u32 * CHUNK_SIZE) & 0xffff).try_into().unwrap(),
                        value: chunk.to_vec(),
                    }),
            );
        }

        records.push(ihex::Record::EndOfFile {});

        let obj = ihex::create_object_file_representation(&records)?;

        Ok(output.write_all(obj.as_bytes())?)
    }
}

impl MemoryWriter for BinWriter {
    fn write_mem(
        &mut self,
        output: &mut dyn std::io::Write,
        _addr: u32, // raw bin writer doesn't care about address
        data: &[u8],
    ) -> Result<(), Box<dyn std::error::Error>> {
        Ok(output.write_all(data)?)
    }
}

impl ObjectFileReader for BinReader {
    fn new(file: &PathBuf) -> Result<Box<dyn ObjectFileReader>, std::io::Error>
    where
        Self: Sized,
    {
        let file = File::open(file)?;
        let filesize = file.metadata()?.len().try_into().unwrap();
        Ok(Box::new(BinReader { file, filesize }))
    }
    fn len(&self) -> usize {
        self.filesize
    }
    fn read_at(
        &mut self,
        buf: &mut Vec<u8>,
        addr: u32,
        size: Option<u32>,
    ) -> Result<(), Box<dyn std::error::Error>> {
        if addr as usize > self.filesize {
            return Err("Can't read past the end of the file!".into());
        }

        let size = std::cmp::min(
            size.unwrap_or(self.filesize as u32 - addr) as usize,
            self.filesize as usize - addr as usize,
        );

        self.file.seek(std::io::SeekFrom::Start(addr.into()))?;
        buf.resize(size as usize, 0);
        Ok(self.file.read_exact(buf)?)
    }
    fn read_first(
        &mut self,
        buf: &mut Vec<u8>,
        size: Option<u32>,
    ) -> Result<(), Box<dyn std::error::Error>> {
        // The casting here is a bit janky...
        let size = std::cmp::min(size.unwrap_or(self.filesize as u32) as usize, self.filesize);
        buf.resize(size as usize, 0);
        Ok(self.file.read_exact(buf)?)
    }
}
